from bellbird.equation import *
from bellbird.operators import *

class Writer:
	def __init__(self, model):
		self.model = model
		self.compiled = False

	def write(self, ln="", t=0, nl=1):
		self.text += t*'\t' + ln + nl*'\n'

	def compile(self, fileName):
		self.name = self.model.name.lower()[0] + "".join(self.model.name.split())[1:]
		self.fileName = fileName

		self.transient = True in [equation.isTransient() for equation in self.model.equations]
		self.transientFields = [ var.name for equation in self.model.equations for term in equation.terms if hasTermWithSubclass(term, TimeDerivative) for var in getTermFields(term) ]
		self.notImplementedTerms = [term for eq in self.model.discretizedEquations for term in eq.terms]

		self.text = ""

		self.writeHeader(t=0)
		self.writeProblemData(0)
		self.writeFields(1)
		self.declareMatrix(1)
		# self.declareProperties(1)
		# self.writeDefinitions(1)
		self.writeUtilFuncs(1)

		self.assembleMatrix(1)
		self.assembleIndependent(1)

		if self.transient:
			self.writeTransientSolverLoop(1)
		else:
			self.writePermanentSolverLoop(1)

		self.writeMainFunction(0)


		self.text = self.text.replace("Δt", "timeStep").replace("'","\"")
		with open(self.fileName, "w") as f:
			f.write(self.text)

		self.compiled = True

	# ------------------------------------------------------------------------------------------------------

	def assembleMatrix(self, t):
		def add(t, i, j, val, nl=1):
			if self.model.sparse:
				self.write(t=t, ln=f"add({i}, {j}, {val})", nl=nl)
			else:
				self.write(t=t, ln=f"matrix[{i}][{j}] += {val}", nl=nl)

		def set(t, i, nl):
			if self.model.sparse:
				self.write(t=t, ln=f"set({i})", nl=nl)
			else:
				self.write(t=t, ln=f"matrix[{i}] = np.zeros({self.model.matrixSize})")
				self.write(t=t, ln=f"matrix[{i}][{i}] = 1.0", nl=nl)

		def writeMatrixSurfaceIntegral(t, term, eqIdx, field, defs, val, vecEq, vecVar, ax, outerFace):
			self.notImplementedTerms.remove(term)

			arrIdx1 = self.model.arranjementDict["eq"][eqIdx]
			arrIdx2 = self.model.arranjementDict["var"][field.name+('_x' if vecVar else '')]
			idxStr1 = ('coord+' if vecEq and ax==0 else ('i+' if vecEq and ax==3 else ''))
			idxStr2 = 'coord+' if vecEq and ax==1 else ('j+' if vecEq and ax==3 else '')

			vT = (2 if ax==3 else 1) if vecEq else 0

			self.write(t=t+0, ln="# " + str(term.arg))
			self.write(t=t+0, ln=f"for face in element.{'faces' if outerFace else 'innerFaces'}:")
			self.write(t=t+1, ln="area = face.area.getCoordinates()[:dimension]")

			for _def in defs:
				self.write(t=t+1, ln=_def)

			if not outerFace:
				self.write(t=t+1, ln="backwardsHandle, forwardHandle = face.getNeighborVerticesHandles()")
			
			if vecEq and ax==3:
				self.write(t=t+1, ln="for i in range(dimension):")
				self.write(t=t+2, ln="for j in range(dimension):")
			elif vecEq:
				self.write(t=t+1, ln="for coord in range(dimension):")


			self.write(t=t+1+vT, ln="for local, vertex in enumerate(element.vertices):")

			if not outerFace:
				add(t=t+2+vT, i=f"backwardsHandle+({idxStr1}{arrIdx1})*numberOfVertices", j=f"vertex.handle+({idxStr2}{arrIdx2})*numberOfVertices", val=f"{val}")
				add(t=t+2+vT, i=f"forwardHandle+({idxStr1}{arrIdx1})*numberOfVertices", j=f"vertex.handle+({idxStr2}{arrIdx2})*numberOfVertices", val=f"-{val}", nl=2)
			else:
				self.write(t=t+2+vT, ln="if type(face) == PyEFVLib.InnerFace:")
				self.write(t=t+3+vT, ln="backwardsHandle, forwardHandle = face.getNeighborVerticesHandles()", nl=2)
				add(t=t+3+vT, i=f"backwardsHandle+({idxStr1}{arrIdx1})*numberOfVertices", j=f"vertex.handle+({idxStr2}{arrIdx2})*numberOfVertices", val=f"{val}")
				add(t=t+3+vT, i=f"forwardHandle+({idxStr1}{arrIdx1})*numberOfVertices", j=f"vertex.handle+({idxStr2}{arrIdx2})*numberOfVertices", val=f"-{val}", nl=2)
				self.write(t=t+2+vT, ln="elif type(face) == PyEFVLib.OuterFace:")
				add(t=t+3+vT, i=f"face.vertex.handle+({idxStr1}{arrIdx1})*numberOfVertices", j=f"vertex.handle+({idxStr2}{arrIdx2})*numberOfVertices", val=f"{val}", nl=2)

		self.write(t=t, ln="def assembleMatrix():")
		self.write(t=t+1, ln="for region in grid.regions:")
		self.declareProperties(t=t+2, region=True)
		self.writeDefinitions(t=t+2)
		self.write(t=t+2, ln="for element in region.elements:")

		def writeMatrixScalarFieldVolumeIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.lhs:
					if term.__class__ == VolumetricIntegral:
						for field in getTermFields(term):
							if field.__class__ == ScalarField and (field == term.arg or ( term.arg.__class__ == Multiplication and field in term.arg.args )):
								self.notImplementedTerms.remove(term)

								termStr = str(term.arg)
								# coeffStr = "vertex.volume * " + termStr.replace(f"* {field.name}", "")
								coeffStr = "vertex.getSubElementVolume(element) * " + termStr.replace(f"* {field.name}", "")

								arrIdx1 = self.model.arranjementDict["eq"][eqIdx]
								idxStr1 = f"+({arrIdx1})*numberOfVertices" if arrIdx1!="0" else ""

								arrIdx2 = self.model.arranjementDict["var"][field.name]
								idxStr2 = f"+({arrIdx2})*numberOfVertices" if arrIdx2!="0" else ""

								self.write(t=t, ln="# " + termStr)
								self.write(t=t, ln="for vertex in element.vertices:")

								add(t=t+1, i=f"vertex.handle{idxStr1}", j=f"vertex.handle{idxStr2}", val=coeffStr, nl=2)
		writeMatrixScalarFieldVolumeIntegral(t+3)

		def writeMatrixGradScalarVolumeIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.lhs:
					if term.__class__ == VolumetricIntegral:
						grad, var = None, None							
						if term.arg.__class__ == Gradient:
							grad = term.arg
						elif term.arg.__class__ == Multiplication:
							gradTerms = [arg for arg in term.arg.args if arg.__class__ == Gradient]
							grad = gradTerms[0] if gradTerms else None
						if grad:
							fields = getTermFields(grad)
							var = fields[0] if (fields and fields[0].__class__ == ScalarField) else None

						if grad and var:
							coeff = Multiplication(*[arg for arg in term.arg.args if arg != grad]) if term.arg.__class__ == Multiplication else "1"
							coeffStr = str(coeff)

							writeMatrixSurfaceIntegral(
								t = t, term = term, eqIdx = eqIdx, field = var,
								defs = [
									"m = element.vertices.size",
									"transposedVoigtArea = getTransposedVoigtArea(face)",
									"shapeFunctions = face.getShapeFunctions()",
									"identityShapeFunctionMatrix = np.array([shapeFunctions, shapeFunctions, shapeFunctions, np.zeros(m), np.zeros(m), np.zeros(m)]) if dimension == 3 else np.array([shapeFunctions, shapeFunctions, np.zeros(m)])",
									f"matrixCoefficients = {coeffStr} * np.matmul(transposedVoigtArea, identityShapeFunctionMatrix)",
								],
								val = "matrixCoefficients[coord][local]",
								vecEq = True, vecVar = False, ax = 0, outerFace = False,
							)
		writeMatrixGradScalarVolumeIntegral(t+3)

		def writeMatrixGradScalarSurfaceIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.lhs:
					if term.__class__ == SurfaceIntegral:
						grad, var = None, None							
						if term.arg.__class__ == Gradient:
							grad = term.arg
						elif term.arg.__class__ == Multiplication:
							gradTerms = [arg for arg in term.arg.args if arg.__class__ == Gradient]
							grad = gradTerms[0] if gradTerms else None
						if grad:
							fields = getTermFields(grad)
							var = fields[0] if (fields and fields[0].__class__ == ScalarField) else None

						if grad and var:
							coeff = Multiplication(*[arg for arg in term.arg.args if arg != grad]) if term.arg.__class__ == Multiplication else "1"
							coeffStr = str(coeff)

							writeMatrixSurfaceIntegral(
								t = t, term = term, eqIdx = eqIdx, field = var,
								defs = [f"matrixCoefficients = {coeffStr} * np.matmul( area.T, face.globalDerivatives )"],
								val = "matrixCoefficients[local]",
								vecEq = False, vecVar = False, ax = None, outerFace = False,
							)
		writeMatrixGradScalarSurfaceIntegral(t+3)

		def writeMatrixSymmetricGradVecSurfaceIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.lhs:
					if term.__class__ == SurfaceIntegral:
						grad, field = None, None
						if term.arg.__class__ == SymmetricGradient:
							grad = term.arg
						elif term.arg.__class__ == Multiplication:
							gradTerms = [arg for arg in term.arg.args if arg.__class__ == SymmetricGradient]
							grad = gradTerms[0] if gradTerms else None

						if grad:
							fields = getTermFields(grad)
							vecField = fields[0] if (fields and fields[0].__class__ == VectorField) else None
							vecField = vecField if (grad.arg == vecField or (grad.arg.__class__ == Multiplication and vecField in grad.arg.args)) else None
						
						if grad and vecField:
							coeff = Multiplication(*[arg for arg in term.arg.args if arg != grad])
							coeffStr = str(coeff)

							writeMatrixSurfaceIntegral(
								t = t, term = term, eqIdx = eqIdx, field = vecField,
								defs = [
									"transposedVoigtArea = getTransposedVoigtArea(face)",
									"voigtGradientOperator = getVoigtGradientOperator(face.globalDerivatives)",
									f"coeff = {coeffStr}",
									"matrixCoefficients = np.einsum('ij,jk,kmn->imn', transposedVoigtArea, coeff, voigtGradientOperator)",
								],
								val = "matrixCoefficients[i][j][local]",
								vecEq = True, vecVar = True, ax = 3, outerFace = False,
							)
		writeMatrixSymmetricGradVecSurfaceIntegral(t+3)

		def writeMatrixVecFieldSurfaceIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.lhs:
					if term.__class__ == SurfaceIntegral:
						for field in getTermFields(term):
							if field.__class__ == VectorField and (field == term.arg or ( term.arg.__class__ == Multiplication and field in term.arg.args )):
								coeff = Multiplication(*[arg for arg in term.arg.args if arg != field]) if (term.arg.__class__ == Multiplication and len(term.arg.args) > 1) else "1"
								coeffStr = str(coeff)

								writeMatrixSurfaceIntegral(
									t = t, term = term, eqIdx = eqIdx, field = field,
									defs = ["shapeFunctions = face.getShapeFunctions()"],
									val = f"{coeffStr} * shapeFunctions[local] * area[coord]",
									vecEq = True, vecVar = True, ax = 1, outerFace = True,
								)
		writeMatrixVecFieldSurfaceIntegral(t+3)

		def writeMatrixDirichletBoundaryConditions(t):
			self.write(t=t+0, ln="# Dirichlet Boundary Conditions")
			for variableName in self.model.variables:
				idxStr = f"+({self.model.arranjementDict['var'][variableName]})*numberOfVertices" if len(self.model.variables) > 1 else ""
				zVar = (variableName[-2:] == "_z")
				a = (1 if zVar else 0)
				if zVar: self.write(t=t, ln="if dimension == 3:")
				self.write(t=t+a+0, ln=f"for bCondition in problemData.dirichletBoundaries['{variableName}']:")
				self.write(t=t+a+1, ln="for vertex in bCondition.boundary.vertices:")
				set(t=t+a+2, i=f"vertex.handle{idxStr}", nl=2)
		writeMatrixDirichletBoundaryConditions(t+1)

		def inverseMatrix(t):
			self.write(t=t, ln="# Inverse Matrix")
			if self.model.sparse:
				self.write(t=t, ln=f"matrix = sparse.csc_matrix( (matrixVals, zip(*matrixCoords)), shape=({self.model.matrixSize}, {self.model.matrixSize}) )")
				self.write(t=t, ln="inverseMatrix = sparse.linalg.inv( matrix )", nl=2)
			else:
				self.write(t=t, ln="inverseMatrix = np.linalg.inv(matrix)")
			self.write(t=t, ln="return inverseMatrix", nl=2)

		inverseMatrix(t+1)

	def assembleIndependent(self, t):
		self.write(t=t, ln="def assembleIndependent():")
		self.write(t=t+1, ln=f"independent = np.zeros({self.model.matrixSize})", nl=2)

		self.write(t=t+1, ln="for region in grid.regions:")
		self.declareProperties(t=t+2, region=True)
		self.writeDefinitions(t=t+2)
		self.write(t=t+2, ln="for element in region.elements:")

		def writeIndependentVolumeIntegral(t):
			for eqIdx, (equation, discretizedEquation) in enumerate(zip(self.model.equations, self.model.discretizedEquations)):
				for term in discretizedEquation.rhs:
					if term.__class__ == VolumetricIntegral and (term.arg.__class__ == Multiplication or hasSubclass(term.arg, Variable)):
						self.notImplementedTerms.remove(term)

						termStr = str(term.arg)
						coeffStr = "vertex.getSubElementVolume(element) * " + termStr

						for field in getTermFields(term):
							if "_old" in field.name:
								fieldName = field.name.replace("_old", "").capitalize()
								coeffStr = coeffStr.replace(f" {field.name}", f" old{fieldName}Field[vertex.handle]")
							else:
								coeffStr = coeffStr.replace(f" {field.name}", f" {field.name}Field[vertex.handle]")

						if equation.order == 0:
							idxStr = f"+({self.model.arranjementDict['eq'][eqIdx]})*numberOfVertices" if len(self.model.variables) > 1 else ""
							self.write(t=t+0, ln="# " + termStr)
							self.write(t=t+0, ln="for vertex in element.vertices:")
							self.write(t=t+1, ln=f"independent[vertex.handle{idxStr}] += " + coeffStr, nl=2)
						elif equation.order == 1:
							arrIdx = self.model.arranjementDict["eq"][eqIdx]
							idxStr = "+" + (f"(coord+{arrIdx})" if arrIdx!="0" else "coord") + "*numberOfVertices"
							for arg in (term.arg.args if hasSubclass(term.arg, Operator) else term.arg):
								if hasSubclass(arg, Vector):
									coeffStr = coeffStr.replace(f" {arg}", f" {arg}[coord]")
							self.write(t=t+0, ln="# " + termStr)
							self.write(t=t+0, ln="for vertex in element.vertices:")
							self.write(t=t+1, ln="for coord in range(dimension):")							
							self.write(t=t+2, ln=f"independent[vertex.handle{idxStr}] += " + coeffStr, nl=2)

					elif term.__class__ == VolumetricIntegral:
						raise Exception(f"Warning: Term {term} is being ignored")
		writeIndependentVolumeIntegral(t+3)

		def writeIndependentVecConstSurfaceIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.rhs:
					if term.__class__ == SurfaceIntegral and not getTermFields(term):
						self.notImplementedTerms.remove(term)

						coeffStr = str(term.arg)

						arrIdx = self.model.arranjementDict["eq"][eqIdx]
						idxStr = f"+({arrIdx})*numberOfVertices" if arrIdx != "0" else ""

						self.write(t=t+0, ln=f"# {coeffStr}")
						self.write(t=t+0, ln="for innerFace in element.innerFaces:")
						self.write(t=t+1, ln="area = innerFace.area.getCoordinates()[:dimension]")
						self.write(t=t+1, ln=f"coefficient = {coeffStr}")
						self.write(t=t+1, ln="coefficient = np.matmul(area.T, coefficient)", nl=2)
						self.write(t=t+1, ln="backwardsHandle, forwardHandle = innerFace.getNeighborVerticesHandles()")
						self.write(t=t+1, ln=f"independent[backwardsHandle{idxStr}] += coefficient")
						self.write(t=t+1, ln=f"independent[forwardHandle{idxStr}]   -= coefficient", nl=2)
		writeIndependentVecConstSurfaceIntegral(t+3)

		def writeIndependentVecFieldSurfaceIntegral(t):
			for eqIdx, discretizedEquation in enumerate(self.model.discretizedEquations):
				for term in discretizedEquation.rhs:
					if term.__class__ == SurfaceIntegral:
						for field in getTermFields(term):
							if field.__class__ == VectorField and (field == term.arg or ( term.arg.__class__ == Multiplication and field in term.arg.args )):
								self.notImplementedTerms.remove(term)

								coeff = Multiplication(*[arg for arg in term.arg.args if arg != field]) if (term.arg.__class__ == Multiplication and len(term.arg.args) > 1) else "1"
								coeffStr = str(coeff)

								fieldName = f"old{field.name.replace('_old','').capitalize()}Field"

								arrIdx1 = self.model.arranjementDict["eq"][eqIdx]
								idxStr1 = f"+({arrIdx1})*numberOfVertices" if arrIdx1 != "0" else ""

								arrIdx2 = self.model.arranjementDict["var"][field.name.replace("_old","_x")]
								idxStr2 = f"+{arrIdx2}" if arrIdx2 != "0" else ""

								self.write(t=t+0, ln=f"# {term.arg}")
								self.write(t=t+0, ln="for face in element.faces:")
								self.write(t=t+1, ln="area = face.area.getCoordinates()[:dimension]")
								self.write(t=t+1, ln="shapeFunctions = face.getShapeFunctions()", nl=2)
								self.write(t=t+1, ln="for coord in range(dimension):")
								self.write(t=t+2, ln="for local, vertex in enumerate(element.vertices):")
								self.write(t=t+3, ln="if type(face) == PyEFVLib.InnerFace:")
								self.write(t=t+4, ln="backwardsHandle, forwardHandle = face.getNeighborVerticesHandles()")
								self.write(t=t+4, ln=f"independent[backwardsHandle{idxStr1}] += {coeffStr} * shapeFunctions[local] * area[coord] * {fieldName}[vertex.handle + (coord{idxStr2})*numberOfVertices]")
								self.write(t=t+4, ln=f"independent[forwardHandle{idxStr1}] -= {coeffStr} * shapeFunctions[local] * area[coord] * {fieldName}[vertex.handle + (coord{idxStr2})*numberOfVertices]", nl=2)
								self.write(t=t+3, ln="elif type(face) == PyEFVLib.OuterFace:")
								self.write(t=t+4, ln=f"independent[face.vertex.handle{idxStr1}] += {coeffStr} * shapeFunctions[local] * area[coord] * {fieldName}[vertex.handle + (coord{idxStr2})*numberOfVertices]", nl=2)
		writeIndependentVecFieldSurfaceIntegral(t+3)

		def writeNeumannBoundaryConditions(t):
			# This signal implies fluxes inwards
			plusVars = [ var.name for equation in self.model.equations for term in equation.lhs if hasTermWithSubclass(term, Divergence) for var in getTermFields(term) ]
			self.write(t=t+0, ln="# Neumann Boundary Condition")
			for variableName in self.model.variables:
				idxStr = f"+({self.model.arranjementDict['var'][variableName]})*numberOfVertices" if len(self.model.variables) > 1 else ""
				signal = "+" if variableName.split("_")[0] in plusVars else "-"
				zVar = (variableName[-2:] == "_z")
				a = (1 if zVar else 0)
				if zVar: self.write(t=t, ln="if dimension == 3:")
				self.write(t=t+a+0, ln=f"for bCondition in problemData.neumannBoundaries['{variableName}']:")
				self.write(t=t+a+1, ln="for facet in bCondition.boundary.facets:")
				self.write(t=t+a+2, ln="for outerFace in facet.outerFaces:")
				self.write(t=t+a+3, ln=f"independent[outerFace.vertex.handle{idxStr}] {signal}= bCondition.getValue(outerFace.handle) * np.linalg.norm(outerFace.area.getCoordinates())", nl=2)
		writeNeumannBoundaryConditions(t+1)

		def writeDirichletBoundaryConditions(t):
			self.write(t=t+0, ln="# Dirichlet Boundary Condition")
			for variableName in self.model.variables:
				idxStr = f"+({self.model.arranjementDict['var'][variableName]})*numberOfVertices" if len(self.model.variables) > 1 else ""
				zVar = (variableName[-2:] == "_z")
				a = (1 if zVar else 0)
				if zVar: self.write(t=t, ln="if dimension == 3:")
				self.write(t=t+a+0, ln=f"for bCondition in problemData.dirichletBoundaries['{variableName}']:")
				self.write(t=t+a+1, ln="for vertex in bCondition.boundary.vertices:")
				self.write(t=t+a+2, ln=f"independent[vertex.handle{idxStr}] = bCondition.getValue(vertex.handle)", nl=2)
		writeDirichletBoundaryConditions(t+1)

		self.write(t=t+1, ln="return independent", nl=2)
	
	# ------------------------------------------------------------------------------------------------------

	def writeTransientSolverLoop(self, t):
		# Written very specifically for problems with only one variable
		self.write(t=t+0, ln="tolerance = problemData.tolerance")
		self.write(t=t+0, ln="difference = 2*tolerance")
		self.write(t=t+0, ln="iteration = 0")
		self.write(t=t+0, ln="currentTime = 0.0")
		self.write(t=t+0, ln="converged = False", nl=2)
		self.write(t=t+0, ln="inverseMatrix = assembleMatrix()", nl=2)
		self.write(t=t+0, ln="while not converged:")
		self.write(t=t+1, ln="independent = assembleIndependent()")
		if len(self.model.variables) == 1:
			if self.model.sparse:
				self.write(t=t+1, ln=f"{self.model.variables[0]}Field = inverseMatrix.dot(independent)", nl=2)
			else:
				self.write(t=t+1, ln=f"{self.model.variables[0]}Field = np.matmul(inverseMatrix, independent)", nl=2)
		else:
			self.write("")
			if self.model.sparse:
				self.write(t=t+1, ln=f"results = inverseMatrix.dot(independent)")
			else:
				self.write(t=t+1, ln=f"results = np.matmul(inverseMatrix, independent)")
			# for i, variableName in enumerate(self.model.variables):
			# 	idxStr = self.model.arranjementDict["var"][variableName]
			# 	zVar = (variableName[-2:] == "_z")
			# 	a = (1 if zVar else 0)
			# 	if zVar: self.write(t=t+1, ln="if dimension == 3:")
			# 	self.write(t=t+a+1, ln=f"{variableName}Field = results[({idxStr})*numberOfVertices:({idxStr}+1)*numberOfVertices]")
			for variableName in self.model.varRadicals:
				if variableName in self.model.variables:
					idxStr = self.model.arranjementDict["var"][variableName]
					self.write(t=t+1, ln=f"{variableName}Field = results[({idxStr})*numberOfVertices:({idxStr}+1)*numberOfVertices]")
				else:
					idxStr = self.model.arranjementDict["var"][variableName+"_x"]
					self.write(t=t+1, ln=f"{variableName}Field = results[({idxStr})*numberOfVertices:({idxStr}+dimension)*numberOfVertices]")
			self.write("")
		fieldDifferences = ['max(abs('+varName+'Field - old'+varName.capitalize()+'Field))' for varName in self.transientFields]
		if len(fieldDifferences) == 1:
			self.write(t=t+1, ln=f"difference = {fieldDifferences[0]}", nl=2)
		else:
			self.write(t=t+1, ln=f"difference = max( {', '.join(fieldDifferences)} )", nl=2)
		for variableName in self.transientFields:
			self.write(t=t+1, ln=f"old{variableName.capitalize()}Field = {variableName}Field.copy()")
		self.write("")
		self.write(t=t+1, ln="currentTime += timeStep")
		self.write(t=t+1, ln="iteration += 1", nl=2)
		for variableName in self.model.variables:
			if "_" in variableName:
				zVar = (variableName[-2:] == "_z")
				a = (1 if zVar else 0)
				if zVar: self.write(t=t+1, ln="if dimension == 3:")
				i = "xyz".index(variableName.split("_")[1])
				self.write(t=t+a+1, ln=f"saver.save('{variableName}', {variableName.split('_')[0]}Field[{i}*numberOfVertices:{i+1}*numberOfVertices], currentTime)")
			else:
				self.write(t=t+1, ln=f"saver.save('{variableName}', {variableName}Field, currentTime)")
		self.write("")
		self.write(t=t+1, ln="print('{:>9}\t{:>14.2e}\t{:>14.2e}\t{:>14.2e}'.format(iteration, currentTime, timeStep, difference))")
		self.write(t=t+1, ln="converged = ( difference <= tolerance ) or ( currentTime >= problemData.finalTime ) or ( iteration >= problemData.maxNumberOfIterations )", nl=2)
		self.write(t=t+1, ln="if iteration >= problemData.maxNumberOfIterations:")
		self.write(t=t+2, ln="break", nl=2)
		self.write(t=t+0, ln="saver.finalize()", nl=2)
	
	def writePermanentSolverLoop(self, t):
		# Written very specifically for problems with only one variable
		self.write(t=t, ln="inverseMatrix = assembleMatrix()")
		self.write(t=t, ln="independent = assembleIndependent()", nl=2)
		if len(self.model.variables) == 1:
			if self.model.sparse:
				self.write(t=t, ln=f"{self.model.variables[0]}Field = inverseMatrix.dot(independent)")
			else:
				self.write(t=t, ln=f"{self.model.variables[0]}Field = np.matmul(inverseMatrix, independent)")
		else:
			if self.model.sparse:
				self.write(t=t, ln=f"results = inverseMatrix.dot(independent)")
			else:
				self.write(t=t, ln=f"results = np.matmul(inverseMatrix, independent)")
			for i, variableName in enumerate(self.model.variables):
				zVar = (variableName[-2:] == "_z")
				a = (1 if zVar else 0)
				if zVar: self.write(t=t, ln="if dimension == 3:")
				self.write(t=t+a, ln=f"{variableName}Field = results[{i}*numberOfVertices:{i+1}*numberOfVertices]")
		self.write(t=t, ln="")
		for variableName in self.model.variables:
			zVar = (variableName[-2:] == "_z")
			a = (1 if zVar else 0)
			if zVar: self.write(t=t, ln="if dimension == 3:")
			self.write(t=t+a, ln=f"saver.save('{variableName}', {variableName}Field, 0.0)")
		self.write(t=t, ln="saver.finalize()", nl=2)

	def writeMainFunction(self, t):
		# def getRegionNames():
		# 	try:
		# 		import sys,os
		# 		sys.path.append(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, 'PyEFVLib'))
		# 		import PyEFVLib
		# 		return PyEFVLib.read(self.model.meshPath).gridData.regionsNames
		# 	except:
		# 		return ["Body"]
		# regionNames = getRegionNames()
		self.write(t=t+0, ln="def main():")
		self.write(t=t+1, ln="problemData = PyEFVLib.ProblemData(")
		self.write(t=t+2, ln=f"meshFilePath = '{self.model.meshPath}',")
		self.write(t=t+2, ln="outputFilePath = 'results',")
		self.write(t=t+2, ln=f"numericalSettings = PyEFVLib.NumericalSettings( timeStep = {self.model.timeStep}, tolerance = {self.model.tolerance}, maxNumberOfIterations = {self.model.maxNumberOfIterations} ),")
		self.write(t=t+2, ln="propertyData = PyEFVLib.PropertyData({")
		for regionName in self.model.properties:
			self.write(t=t+3, ln=f"'{regionName}':")
			self.write(t=t+3, ln="{")
			for propertyName, propertyValue in self.model.properties[regionName].items():
				self.write(t=t+4, ln=f"'{propertyName}': {propertyValue},")
			self.write(t=t+3, ln="},")
		self.write(t=t+2, ln="}),")
		self.write(t=t+2, ln="boundaryConditions = PyEFVLib.BoundaryConditions({")
		for variableName in self.model.variables:
			if [bc for bc in self.model.boundaryConditions if bc.variableName==variableName]:
				self.write(t=t+3, ln=f"'{variableName}': {{")
				for bc in self.model.boundaryConditions:
					if bc.__class__.__name__ == "InitialCondition" and bc.variableName == variableName:
						self.write(t=t+4, ln=f"'InitialValue': {bc.value},")
					if bc.__class__.__name__ == "BoundaryCondition" and bc.variableName == variableName:
						self.write(t=t+4, ln=f"'{bc.boundaryName}': {{ 'condition' : PyEFVLib.{bc.condition}, 'type' : PyEFVLib.Constant, 'value' : {bc.value} }},")
				self.write(t=t+1, ln="\t\t},")
		self.write(t=t+2, ln="}),\n\t)\n")
		self.write(t=t+1, ln=f"{self.name}( problemData )", nl=2)
		self.write(t=t+0, ln="if __name__ == '__main__':")
		self.write(t=t+1, ln="main()")

	# ------------------------------------------------------------------------------------------------------

	def writeHeader(self, t):
		self.write(t=t, ln="import sys,os")
		self.write(t=t, ln="sys.path.append(os.path.join(os.path.dirname(__file__), os.pardir, 'PyEFVLib'))")
		self.write(t=t, ln="import PyEFVLib")
		self.write(t=t, ln="import numpy as np")
		if self.model.sparse:
			self.write(t=t, ln="from scipy import sparse")
			self.write(t=t, ln="import scipy.sparse.linalg")
		self.write("")

	def writeProblemData(self, t):
		self.write(t=t,   ln=f"def {self.name}(problemData):")
		self.write(t=t+1, ln="propertyData 	 = problemData.propertyData")
		if self.transient:
			self.write(t=t+1, ln="timeStep 		 = problemData.timeStep")
		self.write(t=t+1, ln="grid 			 = problemData.grid")
		self.write(t=t+1, ln="numberOfVertices = grid.numberOfVertices")
		self.write(t=t+1, ln="dimension 		 = grid.dimension", nl=2)
		self.write(t=t+1, ln="saver = PyEFVLib.MeshioSaver(grid, problemData.outputFilePath, problemData.libraryPath, extension='xdmf')\n")

	def writeFields(self, t):
		for fieldName in self.model.varRadicals:
			self.write(t=t, ln=f"{fieldName}Field    = np.repeat(0.0, {'' if fieldName in self.model.variables else 'dimension*'}numberOfVertices)")
			if fieldName.split("_")[0] in self.transientFields:
				if fieldName in self.model.variables:
					self.write(t=t, ln=f"old{fieldName.capitalize()}Field = problemData.initialValues['{fieldName}'].copy()", nl=2)
				else:
					self.write(t=t, ln=f"old{fieldName.capitalize()}Field = np.concatenate((problemData.initialValues['{fieldName}_x'], problemData.initialValues['{fieldName}_y'], problemData.initialValues['{fieldName}_z'])) if dimension==3 else np.concatenate((problemData.initialValues['{fieldName}_x'], problemData.initialValues['{fieldName}_y']))", nl=2)

	def declareMatrix(self, t):
		if self.model.sparse:
			self.write(t=t, ln="matrixCoords = []")
			self.write(t=t, ln="matrixVals = []", nl=2)
		else:
			self.write(t=t, ln=f"matrix 		= np.zeros(({self.model.matrixSize}, {self.model.matrixSize}))", nl=2)
		# self.write(t=t, ln=f"independent = np.zeros({self.model.matrixSize})", nl=2)

	def declareProperties(self, t, region=False):
		for propertyName in self.model.propertyNames:
			self.write(t=t, ln=f"{propertyName:<10} = propertyData.get({'region.handle' if region else '0'}, '{propertyName}')")
		self.write("")

	def writeDefinitions(self, t):
		for definition in self.model.definitions:
			self.write(t=t, ln=definition)
		self.write("")

	def writeUtilFuncs(self, t):
		def writeSparseAddFunc(t):
			self.write(t=t+0, ln="def add(i, j, val):")
			self.write(t=t+1, ln="matrixCoords.append((i, j))")
			self.write(t=t+1, ln="matrixVals.append(val)", nl=2)
		def writeSparseSetFunc(t):
			self.write(t=t+0, ln="def set(i):")
			self.write(t=t+1, ln="_matrixCoords = [coord for coord in matrixCoords if coord[0]!=i] + [(i,i)]")
			self.write(t=t+1, ln="_matrixVals   = [val for coord,val in zip(matrixCoords,matrixVals) if coord[0]!=i] + [1.0]")
			self.write(t=t+1, ln="matrixCoords.clear(); matrixCoords.extend(_matrixCoords)")
			self.write(t=t+1, ln="matrixVals.clear(); matrixVals.extend(_matrixVals)", nl=2)
		if self.model.sparse:
			writeSparseAddFunc(t)
			writeSparseSetFunc(t)

		def writeSymmetricUtilFuncs(t):
			self.write(t=t+0, ln="def getTransposedVoigtArea(innerFace):")
			self.write(t=t+1, ln="Sx, Sy, Sz = innerFace.area.getCoordinates()")
			self.write(t=t+1, ln="return np.array([[Sx,0,Sy],[0,Sy,Sx]]) if dimension==2 else np.array([[Sx,0,0,Sy,0,Sz],[0,Sy,0,Sx,Sz,0],[0,0,Sz,0,Sy,Sx]])", nl=2)
			self.write(t=t+0, ln="def getVoigtGradientOperator(globalDerivatives):")
			self.write(t=t+1, ln="if len(globalDerivatives) == 2:")
			self.write(t=t+2, ln="Nx,Ny = globalDerivatives")
			self.write(t=t+2, ln="zero=np.zeros(Nx.size)")
			self.write(t=t+2, ln="return np.array([[Nx,zero],[zero,Ny],[Ny,Nx]])", nl=2)
			self.write(t=t+1, ln="if len(globalDerivatives) == 3:")
			self.write(t=t+2, ln="Nx,Ny,Nz = globalDerivatives")
			self.write(t=t+2, ln="zero=np.zeros(Nx.size)")
			self.write(t=t+2, ln="return np.array([[Nx,zero,zero],[zero,Ny,zero],[zero,zero,Nz],[Ny,Nx,zero],[zero,Nz,Ny],[Nz,zero,Nx]])", nl=2)
		if True in [hasTermWithSubclass(equation.term, SymmetricGradient) for equation in self.model.equations]:
			writeSymmetricUtilFuncs(t)



import os
from bellbird.equation import *
from bellbird.operators import *
from bellbird.writer import Writer

class Model:
	def __init__(self, name, equationsStr, variables, properties, boundaryConditions, definitions=[], meshPath="", sparse=False, timeStep=0.1, tolerance=1e-4, maxNumberOfIterations=300):
		"""
			- Declara as variáveis de entrada necessárias para a resolução do problema
			- Interpreta as equações fornecidas em self.parseEquations()
			- Discretiza as equações fornecidas em self.applyEbEFVM()
		"""
		self.name 				= name
		self.equationsStr 		= equationsStr
		self.variables 			= variables
		self.properties 		= properties
		self.boundaryConditions = boundaryConditions
		self.definitions		= definitions
		self.meshPath 			= meshPath
		self.sparse				= sparse
		self.timeStep			= timeStep
		self.tolerance			= tolerance
		self.maxNumberOfIterations = maxNumberOfIterations
		self.writer = Writer(self)
		self.compiled = False

		self.parseEquations()
		self.applyEbFVM()

	def parseEquations(self):
		"""
			-> Armazena propriedades, definições e equações
				- Armazena os nomes das propriedades passadas em self.properties
				- Armazena os nomes de variáveis definidas
				- Gera os objetos das variáveis definidas, e classifica em Constante, Vetor e Matriz
				- Inicialmente, gera os objetos das equações fornecidas
					- Atualiza as variáveis que são propriedades para constantes
						Obs.: Gostaríamos de passar propriedades variáveis no futuro, contudo equandto
						não é implementado as propriedades são constantes, e podem passar para dentro e 
						para fora de derivadas espaciais e temporais. Portanto deve-se reconsiderar
						a promoção para Constant em updateVariables
					- Atualiza as variáveis que são definições para as variáveis já criadas 
					  e armazenadas em self.definedVars

			-> Keeps track das dimensões para lidar com o arranjo e acoplamento das equações
				- A função _rearrange tem como objetivo re-escrever expressões calculadas envolvendo "numberOfVertices" e "dimension"
				  Ex.: _rearrange( ["dimension", 1] ) = "dimension + 1"
				- Computa os 'radicais' das variáveis, no sentido de remover _x, _y, _z do nome das variáveis
				- Computa as dimensões das variáveis na variável varDims. Se forem passadas variáveis com o sufixo _x sua dimensão será dimension.
					Usamos dimension porque o programa final deve ser capaz de lidar com 2D e 3D.
				- A variável self.arranjementDict["var"] armazena a posição que as variáveis ocupam no vetor de incógnitas
					Ex.: [u_x, u_y, u_z, p] -> self.arranjementDict["var"] = {'u_x':0, 'u_y':1, 'u_z':2, 'p':dimension}
					Ex.: [p, u_x, u_y, u_z] -> self.arranjementDict["var"] = {'p':0, 'u_x':1+0, 'u_y':1+1, 'u_z':1+2}
		"""

		# Armazena propriedades, definições e equações

		self.propertyNames = list(self.properties.values())[0].keys()
		definedNames = [ definition.split(" = ")[0] for definition in self.definitions ]
		self.definedVars = [ Constant(termStr) if not "[" in definition.split(" = ")[1] else (Vector(termStr) if not "[[" in definition.split(" = ")[1].replace(" ", "") else Matrix(termStr)) for termStr, definition in zip(definedNames, self.definitions)]

		self.equations = []
		for equationStr in self.equationsStr:
			equation = Equation(equationStr)
			equation.updatePropertyVars(self.propertyNames)
			equation.updateVariables(self.variables)
			equation.updateDefinitions(self.definedVars)
			self.equations.append(equation)


		# Keeps track das dimensões para lidar com o arranjo e acoplamento das equações

		def _rearrange(terms):
			return str(Equation(f"x={'+'.join(map(str,terms))}").rearrange().term.args[1])

		self.varRadicals = [var.split("_")[0] for var in self.variables]
		self.varRadicals = [var for idx, var in enumerate(self.varRadicals) if not var in self.varRadicals[:idx]]
		varDims = ["1" if var in self.variables else "dimension" for var in self.varRadicals]
		varOffsets = ["0"] + varDims[:-1]

		eqDimensions = ["dimension" if eq.order else "1" for eq in self.equations]
		eqOffsets = ["0"] + [_rearrange(eqDimensions[:idx+1]) for idx,eq_dim in enumerate(eqDimensions[:-1])]

		self.arranjementDict = {"var": dict(), "eq": dict()}
		self.arranjementDict["var"] = {varRadical+("" if varDimension=="1" else "_"+"xyz"[idx]) : _rearrange([varOffset,idx]) for varOffset,varRadical,varDimension in zip(varOffsets,self.varRadicals,varDims) for idx in range({"1":1,"dimension":3}[varDimension])}
		self.arranjementDict["eq"]  = { idx:arr for idx,arr in enumerate(eqOffsets) }

		numberOfVarsStr = _rearrange([varDims.count('1'), f"{varDims.count('dimension')}*dimension"]).replace(" ","")
		self.matrixSize = f"({numberOfVarsStr})*numberOfVertices"

	def applyEbFVM(self):
		self.discretizedEquations = []
		for equationStr in self.equationsStr:
			discretizedEquation = Equation(equationStr)

			discretizedEquation.updatePropertyVars(self.propertyNames)
			discretizedEquation.updateVariables(self.variables)
			discretizedEquation.updateDefinitions(self.definedVars)
			discretizedEquation.integrateInSpace()				# a = b -> iiint(a) = iiint(b)
			discretizedEquation.applyDivergenceTheorem()		# iiint(div(f)) = iint(f)
			discretizedEquation.integrateInTime()				# d/dt(f) = (f-f_old)*(1/Δt)
			discretizedEquation.isolateVariables(self.variables) # x+a=y+c -> x-y=-a+c
			discretizedEquation.unapplyLinearProperty()

			self.discretizedEquations.append(discretizedEquation)

	def compile(self, filename="results.py"):
		self.writer.compile(filename)
		if self.writer.notImplementedTerms:
			print(f"Warning: it was not possible to implement the term(s) {', '.join(map(str,self.writer.notImplementedTerms))}, implement it manually in results.py ")

	def run(self):
		if not (self.writer.compiled and self.writer.fileName=="results.py"):
			self.compile()
		if not self.writer.notImplementedTerms:
			try:
				from results import main
				main()
			except:
				raise Exception("Error while compiling file!")


class BoundaryCondition:
	bcDict = {"DIRICHLET_BOUNDARY_CONDITION":"Dirichlet", "NEUMANN_BOUNDARY_CONDITION":"Neumann"}
	def __init__(self, variableName, boundaryConditionType, boundaryName, value):
		self.variableName 			= variableName
		self.boundaryConditionType 	= boundaryConditionType
		self.boundaryName 			= boundaryName
		self.value 					= value
	@property
	def condition(self):
		return self.bcDict[self.boundaryConditionType]

class InitialCondition:
	def __init__(self, variableName, value):
		self.variableName 			= variableName
		self.value 					= value


from bellbird.operators import *
from bellbird.variables import *

class Equation:
	def __init__(self, equationStr):
		self.equationStr = equationStr
		self.term = parseEquationStr(self.equationStr)

	def __repr__(self):
		return repr(self.term)

	def __str__(self):
		return str(self.term)

	@property
	def lhs(self):
		if self.term.args[0].__class__ == Sum:
			return self.term.args[0].args
		else:
			return [ self.term.args[0] ]

	@property
	def rhs(self):
		if self.term.args[1].__class__ == Sum:
			return self.term.args[1].args
		else:
			return [ self.term.args[1] ]

	@property
	def terms(self):
		return self.lhs + self.rhs

	def applyLinearProperty(self):
		applyLinearProperty(self.term)

	def applyDistributiveProperty(self):
		applyDistributiveProperty(self.term)

	def clusterOperations(self):
		clusterOperations(self.term)

	def unapplyLinearProperty(self):
		unapplyLinearProperty(self.term)
		self.clusterOperations()
		self.applyDistributiveProperty()

	def rearrange(self):
		self.applyDistributiveProperty()	# (a+b)*c = a*c + b*c
		self.clusterOperations()			# times(a, times(b,c)) = times(a, b, c)

		self.applyLinearProperty()			# f(a*(x+y)) = a*f(x) + a*f(y)

		self.applyDistributiveProperty()	# (a+b)*c = a*c + b*c
		self.clusterOperations()			# times(a, times(b,c)) = times(a, b, c)

		self.applyLinearProperty()			# f(a*(x+y)) = a*f(x) + a*f(y)

		return self

	def updatePropertyVars(self, propertyVars):
		updatePropertyVars(self.term, propertyVars)

	def updateVariables(self, variableNames):
		updateVariables(self.term, variableNames)

	def updateDefinitions(self, definedVars):
		updateDefinitions(self.term, definedVars)

	def integrateInTime(self):
		integrateInTime(self.term)
		self.rearrange()

	def integrateInSpace(self):
		self.term.setArgs([ VolumetricIntegral(self.term.args[0]), VolumetricIntegral(self.term.args[1]) ])
		self.rearrange()

	def applyDivergenceTheorem(self):
		applyDivergenceTheorem(self.term)

	def isolateVariables(self, variables):
		lhsVars = [ term for term in self.lhs if 	 containsVariables(term, variables) ]
		lhsInds = [ term for term in self.lhs if not containsVariables(term, variables) ]
		rhsVars = [ term for term in self.rhs if 	 containsVariables(term, variables) ]
		rhsInds = [ term for term in self.rhs if not containsVariables(term, variables) ]

		self.term = Equals( 
			Sum( *lhsVars, Multiplication(minusOne, Sum(*rhsVars)) ) if rhsVars else Sum(*lhsVars),
			Sum( *rhsInds, Multiplication(minusOne, Sum(*lhsInds)) ) if lhsInds else Sum(*rhsInds)
		)

		self.rearrange()

	_order = None
	@property
	def order(self):
		if not self._order:
			self._order = self.getOrder()
		return self._order
	def getOrder(self):
		return getOrder(self.term)

	def isTransient(self):
		return isTransient(self.term)

def hasSubclass(obj, subclass):
	return subclass in obj.__class__.__mro__

def parseEquationStr(eqString):
	# Separate symbols
	for symbol in symbols:
		eqString = eqString.replace(symbol, f" {symbol} ")
	while "  " in eqString:
		eqString = eqString.replace("  ", " ")
	eqString = eqString.replace("d / dt", "d/dt")
	eqString = eqString.replace("grad _s", "grad_s")
	eqComponentsStr = eqString.split()

	# Separate parenthesis
	while "(" in eqComponentsStr:
		i0 = eqComponentsStr.index("(")
		i1 = [(eqComponentsStr[:i].count("(")-eqComponentsStr[:i].count(")") == 1) and (c == ")") for i,c in enumerate(eqComponentsStr)].index(1)
		eqComponentsStr = eqComponentsStr[:i0] + [ eqComponentsStr[i0+1:i1] ] + eqComponentsStr[i1+1:]
	eqComponents = []

	# Create objects
	for idx, compStr in enumerate(eqComponentsStr):
		# Parenthesis
		if type(compStr) == list:
			arg = parseEquationStr( " ".join(compStr) )
			eqComponents.append(arg)

		# Function
		elif compStr in functions:
			comp = functions[compStr]()
			eqComponents.append(comp)

		# Binary Operator
		elif compStr in binaryOperators:
			comp = binaryOperators[compStr]()
			eqComponents.append(comp)

		# Listed Variables
		elif compStr in variablesDict:
			comp = variablesDict[compStr]
			eqComponents.append(comp)

		# Variable
		elif compStr in varTypes:
			name = eqComponentsStr[idx+1][0]
			comp = varTypes[compStr](name)
			eqComponents.append(comp)
			eqComponentsStr.pop(idx+1)
		else:
			comp = Variable(compStr)
			eqComponents.append(comp)

	eqFunctions = [comp for idx, comp in enumerate(eqComponents) if hasSubclass(comp, Function) and not comp.args]
	eqOperators = [comp for idx, comp in enumerate(eqComponents) if hasSubclass(comp, BinaryOperator) and not comp.args]

	eqFunctions.sort( key=lambda func:list(functions.values()).index(func.__class__) )
	eqOperators.sort( key=lambda op:list(binaryOperators.values()).index(op.__class__) )

	for function in eqFunctions:
		idx = eqComponents.index(function)
		arg = eqComponents[idx+1]
		function.setArg(arg)

		eqComponents.pop( idx+1 )

	for operator in eqOperators:
		idx = eqComponents.index(operator)
		args = [ eqComponents[idx-1], eqComponents[idx+1] ]
		operator.setArgs(args)

		eqComponents.pop( idx+1 )
		eqComponents.pop( idx-1 )

	return eqComponents[0]

def updatePropertyVars(term, propertyVars):
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, Variable):
			if arg.name in propertyVars:
				term.args[idx] = Constant(arg.name)

		if hasSubclass(arg, Operator):
			updatePropertyVars(arg, propertyVars)

def updateVariables(term, variableNames):
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, Variable):
			if arg.name in variableNames:
				term.args[idx] = ScalarField(arg.name)
			elif arg.name in [varName.replace("_x","") for varName in variableNames if "_x" in varName]:
				term.args[idx] = VectorField(arg.name)
			# elif arg.__class__ == Variable and arg.name.isnumeric():
			# 	term.args[idx] = Constant(arg.name)

		if hasSubclass(arg, Operator):
			updateVariables(arg, variableNames)	

def updateDefinitions(term, definedVars):
	definedNames = [ var.name for var in definedVars ]
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, Variable):
			if arg.name in definedNames:
				term.args[idx] = definedVars[ definedNames.index(arg.name) ]

		if hasSubclass(arg, Operator):
			updateDefinitions(arg, definedVars)	

def applyLinearProperty(term, flag=1):
	# f(a * X) = a * f(X)
	for idx, arg in enumerate(term.args):
		function = arg.__class__
		if hasSubclass(arg, Function) and arg.linear:
			if hasSubclass(arg.arg, Multiplication):
				constants = [ mArg for mArg in arg.arg.args if hasSubclass(mArg, Constant) ]
				variables = [ mArg for mArg in arg.arg.args if not hasSubclass(mArg, Constant) ]
				varArg    = Multiplication(*variables) if len(variables) > 1 else ( variables[0] if len(variables) == 1 else None )
				
				if constants and varArg:
					term.args[idx] = Multiplication( *constants, function(varArg) )
				elif len(constants) > 1:
					term.args[idx] = Multiplication( *constants )
				elif len(constants) == 1:
					term.args[idx] = constants[0]

			if hasSubclass(arg.arg, Variable) and arg.arg in [zero, zeroVec]:
				term.args[idx] = arg.arg

			for operator in linearOperators:
				if hasSubclass(arg.arg, operator):
					args = arg.arg.args
					term.args[idx] = operator( *[function(arg) for arg in args] )
		if hasSubclass(arg, Operator):
			applyLinearProperty(arg, flag+1)
	if flag > 0:
		applyLinearProperty(term, flag-1)

def unapplyLinearProperty(term):
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, Multiplication):
			for mArg in arg.args:
				if hasSubclass(mArg, Function) and mArg.linear:
					function = mArg.__class__

					variables = [mmArg for mmArg in arg.args if not hasSubclass(mmArg, Constant) and mmArg != mArg]
					constants = [mmArg for mmArg in arg.args if hasSubclass(mmArg, Constant)]

					if constants and variables:
						term.args[idx] = Multiplication(*variables, function(Multiplication(*constants, *mArg.args)))
					elif constants and not variables:
						term.args[idx] = function(Multiplication(*constants, *mArg.args))
					break

		if hasSubclass(arg, Operator):
			unapplyLinearProperty(arg)

def applyDivergenceTheorem(term):
	for idx, arg in enumerate(term.args):
		if arg.__class__ == VolumetricIntegral and arg.arg.__class__ == Divergence:
			divArg = arg.arg.arg
			term.args[idx] = SurfaceIntegral(divArg)

		elif hasSubclass(arg, BinaryOperator):
			applyDivergenceTheorem(arg)

def integrateInTime(term):
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, TimeDerivative) and hasSubclass(arg.arg, Variable):
			var = arg.arg
			varOld = var.__class__(f"{var.name}_old")
			term.args[idx] = Multiplication( Subtraction(var, varOld), overTimeDifferential )

		elif hasSubclass(arg, Operator):
			integrateInTime(arg)

def applyDistributiveProperty(term, flag=0):
	for idx, arg in enumerate(term.args):
		if hasSubclass(arg, Multiplication):
			for sArg in arg.args:
				if hasSubclass(sArg, Sum) or hasSubclass(sArg, Subtraction):
					plus_func = sArg.__class__
					args = [ mArg for mArg in arg.args if not mArg == sArg ]

					term.args[idx] = plus_func(*[Multiplication(*args, ssArg) for ssArg in sArg.args])

					# only one sArg
					break

		if hasSubclass(arg, Operator):
			applyDistributiveProperty(arg, flag+1)

	if flag > 0:
		applyDistributiveProperty(term, flag-1)

def clusterOperations(term, flag=1):
	for idx, arg in enumerate(term.args):
		# Multiplication(A, Multiplication(B, C)) -> Multiplication(A, B, C)
		if hasSubclass(arg, Multiplication):
			args = []
			for mArg in arg.args:
				if hasSubclass(mArg, Multiplication):
					args += mArg.args
				else:
					args.append(mArg)
			arg.args = args

		# Sum(A, Sum(B, C)) -> Sum(A, B, C)
		if hasSubclass(arg, Sum):
			args = []
			for sArg in arg.args:
				if hasSubclass(sArg, Sum):
					args += sArg.args
				else:
					args.append(sArg)
			arg.args = args

		# Subtraction(A, B) -> Sum(A, Multiplication(minusOne, B))
		if hasSubclass(arg, Subtraction):
			term.args[idx] = Sum(arg.args[0], Multiplication(minusOne, arg.args[1]))

		# Multiplication(minusOne, minusOne, A, B, ...) -> Multiplication(A, B, ...)
		if hasSubclass(arg, Multiplication):
			if arg.args.count(minusOne) >= 2:
				arg.args.remove(minusOne)
				arg.args.remove(minusOne)

		# Sum(0, A, B, ...) -> Sum(A, B, ...)
		if hasSubclass(arg, Sum):
			if zero in arg.args:
				arg.args.remove(zero)
			elif zeroVec in arg.args:
				arg.args.remove(zeroVec)

		# Multiplication(1, A, B, ...) -> Multiplication(A, B, ...)
		if hasSubclass(arg, Multiplication):
			if one in arg.args:
				arg.args.remove(one)

		# Multiplication(0, A, B, ...) -> 0
		if hasSubclass(arg, Multiplication):
			if zero in arg.args:
				term.args[idx] = zero

		# Sum() -> 0
		if hasSubclass(arg, Sum):
			if len(arg.args) == 0:
				term.args[idx] = zero

		# Multiplication() -> 1
		if hasSubclass(arg, Multiplication):
			if len(arg.args) == 0:
				term.args[idx] = one

		# binaryOperator(A) -> A
		if hasSubclass(arg, BinaryOperator):
			if len(arg.args) == 1:
				term.args[idx] = arg.args[0]

		# d/dX( const(A) ) -> 0
		if arg.__class__ in [TimeDerivative, Gradient]:
			if hasSubclass(arg.arg, Constant):
				term.args[idx] = zero

		# vec(0) -> zeroVec
		if hasSubclass(arg, Vector):
			if arg.name == "0":
				term.args[idx] = zeroVec

		# d/dt(div(X)) -> div(d/dt(X))
		# Done to use the divergence theorem
		if hasSubclass(arg, TimeDerivative):
			if hasSubclass(arg.arg, Divergence):
				term.args[idx] = Divergence(TimeDerivative( arg.arg.arg ))

		# # 1+1 -> 2
		# if hasSubclass(arg, BinaryOperator):
		# 	numericArgs = [oArg for oArg in arg.args if hasSubclass(oArg, Variable) and oArg.name.isnumeric()]
		# 	if len(numericArgs) >= 2:
		# 		numericalResult = eval(str(arg.__class__(*numericArgs)))
		# 		arg.args = [Constant(str(numericalResult))] + [oArg for oArg in arg.args if not oArg in numericArgs]

		# # a + a + a -> 3*a
		# if hasSubclass(arg, Operator):
		# 	repeatedArgs = [oArg for oArg in arg.args if list(map(str,arg.args)).count(str(oArg))>1]
		# 	repeatedArgs = [oArg for idx,oArg in enumerate(repeatedArgs) if not str(oArg) in map(str,repeatedArgs[:idx])]
		# 	for rArg in repeatedArgs:
		# 		numberOfTimes = list(map(str,arg.args)).count(str(rArg))
		# 		arg.args = [Multiplication(Constant(str(numberOfTimes)), rArg)] + [oArg for oArg in arg.args if not str(oArg)==str(rArg)]

		# # n*x*A + k*y*A + B -> (n*x+k*y) * A + B
		# if hasSubclass(arg, Sum):
		# 	variables = [var for var in getTermVars(arg) if not var.name.isnumeric()]
		# 	variables = [var for idx,var in enumerate(variables) if var.name not in map(str,variables[:idx])]
		# 	for var in variables:
		# 		sharingTerms = [sArg for sArg in arg.args if hasSubclass(sArg,Multiplication) and var.name in map(str,sArg.args)]
		# 		if len(sharingTerms)>1:
		# 			for sTerm in sharingTerms:
		# 				sTerm.args = [stArg for stArg in sTerm.args if str(var)!=str(stArg)]
		# 				arg.args.remove(sTerm)
		# 			arg.args.append( Multiplication(Sum(*sharingTerms), var) )

		if hasSubclass(arg, Operator):
			clusterOperations(arg, flag+1)

	if flag > 0:
		clusterOperations(term, flag-1)
		
def containsVariables(term, variables):
	if hasSubclass(term, Variable):
		return term.name in variables + [varName.replace("_x", "") for varName in variables if "_x" in varName]

	elif hasSubclass(term, Operator):
		for arg in term.args:
			if containsVariables(arg, variables):
				return True

		return False

def getTermFields(term):
	if hasSubclass(term, Field):
		return [term]
	
	fields = []
	if hasSubclass(term, Operator):
		for arg in term.args:
			fields += getTermFields(arg)

	return fields

def getTermVars(term):
	if hasSubclass(term, Variable):
		return [term]
	
	variables = []
	if hasSubclass(term, Operator):
		for arg in term.args:
			variables += getTermVars(arg)

	return variables

def hasTermWithSubclass(term, subclass):
	if hasSubclass(term, subclass):
		return True
	elif hasSubclass(term, Operator):
		for arg in term.args:
			if hasTermWithSubclass(arg, subclass):
				return True
	return False

def getOrder(term):
	if hasSubclass(term, Scalar):
		return 0
	elif hasSubclass(term, Vector):
		return 1
	elif hasSubclass(term, Matrix):
		return 2
	elif hasSubclass(term, Tensor):
		return 3
	elif hasSubclass(term, Function):
		if term.__class__ in [TimeDerivative, TimeIntegral, VolumetricIntegral, SurfaceIntegral, VolumetricSummation, SurfaceSummation]:
			return getOrder(term.arg)
		elif term.__class__ == Divergence:
			return getOrder(term.arg) - 1
		elif term.__class__ in [Gradient, SymmetricGradient]:
			return getOrder(term.arg) + 1

	elif hasSubclass(term, BinaryOperator):
		if term.__class__ in [Sum, Subtraction, Equals]:
			argsOrders = [ getOrder(arg) for arg in term.args ]
			order = argsOrders[0]
			if [argOrder for argOrder in argsOrders if argOrder != order ]:
				raise Exception(f"Invalid operation {term}")
			return order

		elif term.__class__ == Multiplication:
			argsOrders = [ getOrder(arg) for arg in term.args ]
			order = argsOrders[0]

			for argOrder in argsOrders[1:]:
				if order == 0:
					order = argOrder
				elif order == 1 and argOrder == 1:
					order = 0
				elif order == 2 and argOrder == 1:
					order = 1
			return order

		elif term.__class__ == Division:
			argsOrders = [ getOrder(arg) for arg in term.args ]
			if argsOrders[0] != argsOrders[1]:
				raise Exception(f"Invalid term {term}")

			return argsOrders[0]

		elif term.__class__ == DotProduct:
			return sum([ getOrder(arg) for arg in term.args ]) - 2

		elif term.__class__ == CrossProduct:
			# May need revision
			return getOrder(term.args[0])

	if term.__class__ == Variable:
		raise Exception(f"Variable {term} was not declared as a property")
	else:
		raise Exception(f"Could not find order of the term {term}")

def isTransient(term):
	if hasSubclass(term, TimeDerivative):
		return True

	elif hasSubclass(term, Operator):
		for arg in term.args:
			if isTransient(arg):
				return True

	return False


from .variables import *
from .operators import *
from .equation import *
from .model import *
from .writer import *

Dirichlet = "DIRICHLET_BOUNDARY_CONDITION"
Neumann   = "NEUMANN_BOUNDARY_CONDITION"


class BoundaryCondition:
	bcDict = {"DIRICHLET_BOUNDARY_CONDITION":"Dirichlet", "NEUMANN_BOUNDARY_CONDITION":"Neumann"}
	conditionType = "BOUNDARY_CONDITION"
	def __init__(self, variableName, boundaryConditionType, boundaryName, value):
		self.variableName 			= variableName
		self.boundaryConditionType 	= boundaryConditionType
		self.boundaryName 			= boundaryName
		self.value 					= value
	@property
	def condition(self):
		return self.bcDict[self.boundaryConditionType]

class InitialCondition:
	def __init__(self, variableName, value):
		self.variableName 			= variableName
		self.value 					= value


# --------- Operators ---------
class Operator:
	def __init__(self, *args):
		self.args = list(args)

	def setArgs(self, args):
		self.args = args

	def setArg(self, arg):
		self.args = [ arg ]

class Function(Operator):
	def __repr__(self):
		if self.args:
			return f"{self.func}({repr(self.args[0])})"
		else:
			return f"{self.func}()"

	def __str__(self):
		return f"{self.func}({self.args[0]})"

	@property
	def arg(self):
		return self.args[0]
	

class BinaryOperator(Operator):
	def __repr__(self):
		# return f" {self.symbol} ".join(map(repr, self.args))
		return f"{self.func}({', '.join(map(repr, self.args))})"

	def __str__(self):
		return f" {self.symbol} ".join(map(str, self.args))


# --------- Functions --------- 
class Divergence(Function):
	func = "div"
	linear = True

class Gradient(Function):
	func = "grad"
	linear = True

class SymmetricGradient(Function):
	# (1/2) * ( grad(u) + grad(u)^T )
	func = "grad_s"
	linear = True

class TimeDerivative(Function):
	func = "d/dt"
	linear = True

class TimeIntegral(Function):
	func = "intT"
	linear = True

class VolumetricIntegral(Function):
	func = "iiint"
	linear = True

class SurfaceIntegral(Function):
	func = "iint"
	linear = True

class VolumetricSummation(Function):
	func = "summ3D"
	linear = True

class SurfaceSummation(Function):
	func = "summ2D"
	linear = True

# --------- BinaryOperator --------- 
class Equals(BinaryOperator):
	symbol = "="
	func   = "equals"
	commutative = True

class Sum(BinaryOperator):
	symbol = "+"
	func   = "sum"
	commutative = True

class Subtraction(BinaryOperator):
	symbol = "-"
	func   = "subtraction"
	commutative = False

class Multiplication(BinaryOperator):
	symbol = "*"
	func   = "multiplication"
	commutative = True

class Division(BinaryOperator):
	symbol = "/"
	func   = "division"
	commutative = False

class DotProduct(BinaryOperator):
	symbol = "·"
	func   = "dotProduct"
	commutative = True

class CrossProduct(BinaryOperator):
	symbol = "×"
	func   = "crossProduct"
	commutative = False


# --------- Dicts ---------

functions = {
	"div"	 : Divergence,
	"grad"	 : Gradient,
	"grad_s" : SymmetricGradient,
	"d/dt"	 : TimeDerivative,
	"intT"	 : TimeIntegral,
	"iiint"	 : VolumetricIntegral,
	"iint"	 : SurfaceIntegral,
	"summ3D" : VolumetricSummation,
	"summ2D" : SurfaceSummation,
}

binaryOperators = {
	"/"		: Division,
	"*"		: Multiplication,
	"·"		: DotProduct,
	"×"		: CrossProduct,
	"-"		: Subtraction,
	"+"		: Sum,
	"="		: Equals,	
}

linearOperators = [Sum, Subtraction, Equals]

other = ["(", ")"]

symbols = list(functions.keys()) + list(binaryOperators.keys()) + other


class Variable:
	symbol = "var"
	def __init__(self, name=None):
		self.name = name

	def __repr__(self):
		return self.name
		# return f"{self.symbol}({self.name})"

	def __str__(self):
		return self.name

	def setVarName(self, name):
		self.name = name

class Scalar(Variable):
	symbol = "scalar"

class Vector(Variable):
	symbol = "vec"

class Matrix(Variable):
	symbol = "mat"

class Tensor(Variable):
	symbol = "tens"

class Constant(Scalar):
	symbol = "const"

class Field(Variable):
	symbol = "field"

class ScalarField(Scalar, Field):
	symbol = "scalarF"

class VectorField(Vector, Field):
	symbol = "vecF"

varTypes = {
	"var"	:	Variable,
	"scalar": 	Scalar,
	"vec"	:	Vector,
	"mat"	:	Matrix,
	"tens"	:	Tensor,
	"const"	:	Constant,
	"field"	:	Field,
	"scalarF":	ScalarField,
	"vecF"	:	VectorField,
}

timeDifferential 		= Scalar("Δt")
overTimeDifferential 	= Scalar("(1/Δt)")
minusOne				= Constant("(-1)")
zero					= Constant("0")
one 					= Constant("1")
zeroVec					= Vector("0")
identityTensor			= Tensor("I")

variablesDict = {
	"timeDifferential"		: timeDifferential,
	"dt"					: timeDifferential,
	"Δt"					: timeDifferential,
	"overTimeDifferential"	: overTimeDifferential,
	"(1/Δt)"				: overTimeDifferential,
	"(1/dt)"				: overTimeDifferential,
	"minusOne"				: minusOne,
	"(-1)"					: minusOne,
	"zero"					: zero,
	"0"						: zero,
	"one"					: one,
	"1"						: one,
	"zeroVec"				: zeroVec,
	"I"						: identityTensor,
}


from setuptools import setup

with open("README.md", "r") as f:
	long_description = f.read()

setup(
	name="bellbird",
	version="0.0.1",
	description="Library that parses differential equations (PDEs) strings and automates the creation of scripts for PyEFVLib using the element-based finite volume method. ",
	py_modules=["bellbird/__init__", "bellbird/variables", "bellbird/operators", "bellbird/equation", "bellbird/model", "bellbird/writer"],
	classifiers=[
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent"
	],
	long_description=long_description,
	long_description_content_type="text/markdown",
	install_requires=[
		"numpy ~= 1.19.2",
		"PyEFVLib ~= 1.0.1",
	],
	url="https://github.com/GustavoExel/bellbird",
	author="Gustavo Exel",
	author_email="gustavoexelgpe@gmail.com",

)